# gen_read_callsite_file.py
from __future__ import absolute_import

from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin
from .verify_utils import VERIFY_PBLOCK_USE_PART, VERIFY_PBLOCK_DECL_PART, VERIFY_PBLOCK_EXEC_PART, \
    VERIFY_PBLOCK_CONTAINS_PART, VERIFY_PBLOCK_SUBP_PART, VERIFY_PBLOCK_EXTERNS, VERIFY_PBLOCK_LOCALS, \
    VERIFY_PBLOCK_INIT

class Verify_K_Callsite_File(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('parentblock_stmt'), None, self.create_parentblock_parts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('topblock_stmt'), None, self.create_topblock_parts)

    def create_topblock_parts(self, node):

        #attrs = {'name': 'kgen_utils_mod', 'isonly': True, 'items':['check_t', 'kgen_init_check', 'kgen_tolerance', \
        #    'kgen_minvalue', 'CHECK_IDENTICAL', 'CHECK_IN_TOL', 'CHECK_OUT_TOL']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        prenode = getinfo('blocknode_aftercallsite_main')
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            prenode, None, self.create_verification_parts)

        namedpart_create_subpart(prenode, VERIFY_PBLOCK_INIT, EXEC_PART)
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_INIT, '')
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_INIT, 'verify init')

        namedpart_create_subpart(prenode, VERIFY_PBLOCK_EXTERNS, EXEC_PART)
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_EXTERNS, '')
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_EXTERNS, 'extern verify variables')

        namedpart_create_subpart(prenode, VERIFY_PBLOCK_LOCALS, EXEC_PART)
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_LOCALS, '')
        namedpart_append_comment(prenode.kgen_kernel_id, VERIFY_PBLOCK_LOCALS, 'local verify variables')

    def create_verification_parts(self, node):

        attrs = {'designator': 'kgen_init_verify', 'items': ['tolerance=%s'%getinfo('verify_tol'), \
            'minvalue=%s'%getinfo('verify_minval'), 'verboseLevel=%s'%getinfo('verbose_level')]}
        namedpart_append_genknode(node.kgen_kernel_id, VERIFY_PBLOCK_INIT, statements.Call, attrs=attrs)

        attrs = {'designator': 'kgen_init_check', 'items': ['check_status', 'rank=kgen_mpirank']}
        namedpart_append_genknode(node.kgen_kernel_id, VERIFY_PBLOCK_INIT, statements.Call, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        # verification statistics
        attrs = {'expr': 'kgen_verboseLevel > 0'}
        ifstatobj = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(ifstatobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"Number of output variables: "','check_status%numTotal']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Number of identical variables: "','check_status%numIdentical']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Number of non-identical variables within tolerance: "','check_status%numInTol']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Number of non-identical variables out of tolerance: "','check_status%numOutTol']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Tolerance: "','kgen_tolerance']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        # verification result
        attrs = {'expr': 'check_status%numOutTol > 0'}
        ifobj = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(ifobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"Verification FAILED with" // TRIM(ADJUSTL(kgen_filepath))']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)
        
        attrs = {'variable': 'check_status%Passed', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_isverified', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifobj, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_genknode(ifobj, EXEC_PART, statements.Else, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(ifobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"Verification PASSED with " // TRIM(ADJUSTL(kgen_filepath))']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)
        
        attrs = {'variable': 'check_status%Passed', 'sign': '=', 'expr': '.TRUE.'}
        part_append_genknode(ifobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_isverified', 'sign': '=', 'expr': '.TRUE.'}
        part_append_genknode(ifobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'check_status%rank == 0'}
        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        if getinfo('add_mpi_frame'):
            part_append_comment(node, EXEC_PART, '')
            part_append_comment(node, EXEC_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, EXEC_PART, 'call mpi_barrier(mpi_comm_world, kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')
            part_append_comment(node, EXEC_PART, '')

    def create_parentblock_parts(self, node):

        namedpart_link_part(node, VERIFY_PBLOCK_USE_PART, USE_PART)
        namedpart_link_part(node, VERIFY_PBLOCK_DECL_PART, DECL_PART)
        namedpart_link_part(node, VERIFY_PBLOCK_EXEC_PART, EXEC_PART)
        namedpart_link_part(node, VERIFY_PBLOCK_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, VERIFY_PBLOCK_SUBP_PART, SUBP_PART)

        attrs = {'type_spec': 'TYPE', 'selector':(None, 'check_t'), 'entity_decls': ['check_status']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Type, attrs=attrs)



from microapp import Project

# TODO: buildmodel and runmodell will go to langlab project

from fortlab.scanner import (MicroappBuildScanner, MicroappRunScanner,
                             MicroappModelCombiner)
from fortlab.scanner.compile import FortranCompilerOption
from fortlab.scanner.timing import (FortranTimingGenerator,
            FortranTimingCollector, FortranTimingCombiner)
from fortlab.resolver import FortranNameResolver
from fortlab.kernel import FortranKernelGenerator
from fortlab.state import FortranStateGenerator
from fortlab.vargen import FortranVariableAnalyzer

class Fortlab(Project):
    _name_ = "fortlab"
    _version_ = "0.1.16"
    _description_ = "Fortran Kernel and Analysis Framework"
    _long_description_ = "Tools for Analysis of Fortran Application and Source code"
    _author_ = "Youngsung Kim"
    _author_email_ = "youngsung.kim.act2@gmail.com"
    _url_ = "https://github.com/grnydawn/fortlab"
    _builtin_apps_ = [MicroappBuildScanner, MicroappRunScanner,
                      FortranCompilerOption, FortranNameResolver,
                      FortranTimingGenerator, FortranKernelGenerator,
                      FortranStateGenerator, MicroappModelCombiner,
                      FortranTimingCollector, FortranTimingCombiner,
                      FortranVariableAnalyzer]
    _requires_ = ["dict2json>=0.1.2"]

    def __init__(self):
        self.add_argument("--test", help="test argument")



# gen_read_callsite_file.py

import os
from fortlab.resolver import block_statements
from fortlab.kgplugin import Kgen_Plugin

class Gen_K_Files(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            block_statements.BeginSource, None, self.add_file_header)

    def add_file_header(self, node):
        from time import strftime

        part_insert_comment(node, UNIT_PART, 0, 'KGEN-generated Fortran source file')
        part_insert_comment(node, UNIT_PART, 1, '')
        part_insert_comment(node, UNIT_PART, 2, 'Generated at : %s'%strftime("%Y-%m-%d %H:%M:%S"))
        part_insert_comment(node, UNIT_PART, 3, 'KGEN version : %s'%getinfo('kgen_version'))
        part_insert_comment(node, UNIT_PART, 4, '')
        #part_insert_comment(node, UNIT_PART, 5, 'KERNEL LICENSE      : Please see KERNEL_LICENSE.txt for details.')



"""
Various utility functions.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['split_comma', 'specs_split_comma',
           'ParseError','AnalyzeError',
           'get_module_file','parse_bind','parse_result','is_name','parse_array_spec',
           'CHAR_BIT','str2stmt',
           'entity_split_comma', 'classes'] # KGEN addition
           #'classes'] # KGEN deletion

import re
import abc
import os, glob
import sys
import io
import traceback

class ParseError(Exception):
    pass

class AnalyzeError(Exception):
    pass

is_name = re.compile(r'^[a-z_]\w*$',re.I).match
name_re = re.compile(r'[a-z_]\w*',re.I).match
is_entity_decl = re.compile(r'^[a-z_]\w*',re.I).match
is_int_literal_constant = re.compile(r'^\d+(_\w+|)$').match
#module_file_extensions = ['.f', '.f90', '.f95', '.f03', '.f08'] # KGEN deletion
module_file_extensions = ['.f', '.f90', '.f95', '.f03', '.f08', '.F', '.F90', '.F95', '.F03', '.F08'] # KGEN addition

# start of KGEN addition
def entity_split_comma(line, item = None, comma=',', keep_empty=False):
    items = []
    if item is None:
        s = ''
        depth = 0
        for ch in line:
            if ch == '[':
                depth += 1
            elif ch == ']':
                depth -= 1

            if ch == comma:
                if depth == 0:
                    if not s and not keep_empty: continue
                    items.append(s.strip())
                    s = ''
                elif depth < 0:
                    raise
                else:
                    s += ch
            else:
                s += ch
        if s: items.append(s.strip())
        return items

    if not line.strip(): # we may have blank space so strip the line
        return []

    newitem = item.copy(line, True)
    apply_map = newitem.apply_map
    s = ''
    depth = 0
    for ch in newitem.get_line().strip():
        if ch == '[':
            depth += 1
        elif ch == ']':
            depth -= 1

        if ch == comma:
            if depth == 0:
                s = apply_map(s).strip()
                if not s and not keep_empty: continue
                items.append(s)
                s = ''
            elif depth < 0:
                raise
            else:
                s += ch
        else:
            s += ch
    s = apply_map(s).strip()
    if s: items.append(s)
    return items
# end of KGEN addition

def split_comma(line, item = None, comma=',', keep_empty=False):
    items = []
    if item is None:
        for s in line.split(comma):
            s = s.strip()
            if not s and not keep_empty: continue
            items.append(s)
        return items
    if not line.strip(): # we may have blank space so strip the line
        return []
    newitem = item.copy(line, True)
    apply_map = newitem.apply_map
    for s in newitem.get_line().split(comma):
        s = apply_map(s).strip()
        if not s and not keep_empty: continue
        items.append(s)
    return items

def parse_array_spec(line, item = None):
    items = []
    for spec in split_comma(line, item):
        items.append(tuple(split_comma(spec, item, comma=':', keep_empty=True)))
    return items

def specs_split_comma(line, item = None, upper=False):
    specs0 = split_comma(line, item)
    specs = []
    for spec in specs0:
        i = spec.find('=')
        if i!=-1:
            kw = spec[:i].strip().upper()
            v  = spec[i+1:].strip()
            specs.append('%s = %s' % (kw, v))
        else:
            if upper:
                spec = spec.upper()
            specs.append(spec)
    return specs

def parse_bind(line, item = None):
    if not line.lower().startswith('bind'):
        return None, line
    if item is not None:
        newitem = item.copy(line, apply_map=True)
        newline = newitem.get_line()
    else:
        newitem = None
    newline = newline[4:].lstrip()
    i = newline.find(')')
    assert i!=-1,str(newline)
    args = []
    for a in specs_split_comma(newline[1:i].strip(), newitem, upper=True):
        args.append(a)
    rest = newline[i+1:].lstrip()
    if item is not None:
        rest = newitem.apply_map(rest)
    return args, rest

def parse_result(line, item = None):
    if not line.lower().startswith('result'):
        return None, line
    line = line[6:].lstrip()
    i = line.find(')')
    assert i != -1,str(line)
    name = line[1:i].strip()
    assert is_name(name),str(name)
    return name, line[i+1:].lstrip()

def filter_stmts(content, classes):
    """ Pop and return classes instances from content.
    """
    stmts = []
    indices = []
    for i in range(len(content)):
        stmt = content[i]
        if isinstance(stmt, classes):
            stmts.append(stmt)
            indices.append(i)
    indices.reverse()
    for i in indices:
        del content[i]
    return stmts


def get_module_files(directory, _cache={}):
    if directory in _cache:
        return _cache[directory]
    module_line = re.compile(r'(\A|^)module\s+(?P<name>\w+)\s*(!.*|)$',re.I | re.M)
    d = {}
    files = []
    for ext in module_file_extensions:
        files += glob.glob(os.path.join(directory,'*'+ext))
    for fn in files:
        f = io.open(fn,'r', encoding="utf-8")
        for name in module_line.findall(f.read()):
            name = name[1]
            if name in d:
                print(d[name],'already defines',name)
                continue
            d[name] = fn
    _cache[directory] = d
    return d

def get_module_file(name, directory, _cache={}):
    fn = _cache.get(name, None)
    if fn is not None:
        return fn
    if name.endswith('_module'):
        for ext in module_file_extensions:
            f1 = os.path.join(directory,name[:-7]+ext)
            if os.path.isfile(f1):
                _cache[name] = fn
                return f1
    files = []
    for ext in module_file_extensions:
        files += glob.glob(os.path.join(directory,'*'+ext))
    for fn in files:
        if module_in_file(name, fn):
            _cache[name] = fn
            return fn
    return

def module_in_file(name, filename):
    name = name.lower()
    pattern = re.compile(r'\s*module\s+(?P<name>[a-z]\w*)', re.I).match

    # KGEN added latin-1 encoder

    try:
        with io.open(filename,'r', encoding="utf-8") as f:
            lines = f.readlines()

    except UnicodeDecodeError as err:
        with io.open(filename,'r', encoding="latin-1") as f:
            lines = f.readlines()

    for line in lines:
        m = pattern(line)
        if m and m.group('name').lower()==name:
            #f.close()
            return filename
    #f.close()


def str2stmt(string, isfree=True, isstrict=False):
    """ Convert Fortran code to Statement tree.
    """
    from readfortran import Line, FortranStringReader
    from parsefortran import FortranParser
    reader = FortranStringReader(string, isfree, isstrict)
    parser = FortranParser(reader)
    parser.parse()
    parser.analyze()
    block = parser.block
    while len(block.content)==1:
        block = block.content[0]
    return block

def get_char_bit():
    # start of KGEN deletion
#    import numpy
#    one = numpy.ubyte(1)
#    two = numpy.ubyte(2)
#    n = numpy.ubyte(2)
#    i = 1
#    while n>=two:
#        n <<= one
#        i += 1
#    return i
    # end of KGEN deletion

    # start of KGEN addition
    try:
        for i in range(16):
            chr(1<<i)
    except:
        return i
    # end of KGEN addition

CHAR_BIT = get_char_bit()

def show_item_on_failure(func, _exception_depth=[0]):
    """
    Decorator for analyze methods.
    """
    def new_func(self):
        try:
            func(self)
        except AnalyzeError as msg:
            clsname = self.__class__.__name__
            self.error('%s.analyze error: %s' % (clsname,msg))
            traceback.print_exc()
        except ParseError as msg:
            self.error('parse error: %s' % (msg))
        except Exception as msg:
            _exception_depth[0] += 1
            if _exception_depth[0]==1:
                self.error('exception triggered here: %s %s' % (Exception, msg))
            raise
        _exception_depth[0] = 0
    return new_func

_classes_cache = {}

class meta_classes(type):
    """ Meta class for ``classes``.
    """

    __abstractmethods__ = False
    def __getattr__(self, name):
        # Expose created classes only as attributes to ``classes`` type.
        cls = _classes_cache.get(name)
        if cls is None:
            raise AttributeError('instance does not have attribute %r' % (name))
        return cls

class classes(type):
    pass

#class classes(type):
##class classes(metaclass=meta_classes):
#    """Make classes available as attributes of this class.
#
#    To add a class to the attributes list, one must use::
#
#      __metaclass__ = classes
#
#    in the definition of the class.
#
#    In addition, apply the following tasks:
#
#    * decorate analyze methods with show_item_on_failure
#    """
#
#    #__metaclass__ = meta_classes
#
#    __abstractmethods__ = False
#    def __getattr__(self, name):
#        if name != "a": import pdb; pdb.set_trace()
#        # Expose created classes only as attributes to ``classes`` type.
#        cls = _classes_cache.get(name)
#        if cls is None:
#            raise AttributeError('instance does not have attribute %r' % (name))
#        return cls
#
#    def __new__(metacls, name, bases, dict):
#        if 'analyze' in dict:
#            dict['analyze'] =  show_item_on_failure(dict['analyze'])
#        cls = type.__new__(metacls, name, bases, dict)
#        _classes_cache[name] = cls
#        return cls
#




# gencore_write_subr.py
from __future__ import absolute_import

from fortlab.resolver import statements, block_statements, typedecl_statements
from .gencore_utils import kernel_gencore_contains, state_gencore_contains, get_dtype_writename, get_dtype_readname, \
    gen_read_istrue, gen_write_istrue, check_class_derived

def create_read_subr(subrname, entity_name, parent, var, stmt, allocate=False, ename_prefix=''):

    checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname

    is_class_derived = check_class_derived(stmt)

    if not part_has_node(parent, SUBP_PART, checks):

        checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
        if not parent in kernel_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
            part_append_comment(parent, CONTAINS_PART, '')
            part_append_genknode(parent, CONTAINS_PART, statements.Contains)
            part_append_comment(parent, CONTAINS_PART, '')
            kernel_gencore_contains.append(parent)

        part_append_comment(parent, SUBP_PART, 'read state subroutine for %s'%subrname)
        # TODO: may add INTENT type after var
        attrs = {'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
        subrobj = part_append_genknode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(parent, SUBP_PART, '')

        # variable A
        #import pdb; pdb.set_trace()
        attrspec = ['INTENT(INOUT)']
        if var.is_pointer(): attrspec.append('POINTER')
        if var.is_allocatable() or allocate: attrspec.append('ALLOCATABLE')
        if var.is_array(): attrspec.append('DIMENSION(%s)'% ','.join(':'*var.rank))
        attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, 'selector':stmt.selector, 'entity_decls': ['var']}
        part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

        # kgen_unit
        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        # printname
        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # printvar
        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # kgen_istrue
        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
        part_append_genknode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

        # array index A
        if var.is_array():
            attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'idx%d'%(d+1) for d in range(var.rank) ]}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['DIMENSION(2,%d)'%var.rank], 'entity_decls': [ 'kgen_bound' ]}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        part_append_comment(subrobj, DECL_PART, '')

        pobj = gen_read_istrue(subrobj, var, 'var', allocate=allocate)

        if var.is_array():
            bound_args = []
            for dim in range(var.rank):
                attrs = {'items': ['kgen_bound(1, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                attrs = {'items': ['kgen_bound(2, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                #bound_args.append('kgen_bound(2,%d)-kgen_bound(1,%d)+1'%(dim+1, dim+1))
                bound_args.append('kgen_bound(1,%d):kgen_bound(2,%d)'%(dim+1, dim+1))

            if var.is_allocatable() or var.is_pointer() or allocate:
                attrs = {'items': ['var(%s)'%', '.join(bound_args)]}
                part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

            if stmt.is_derived() or is_class_derived:
                indexes = [ 'idx%d'%(d+1) for d in range(var.rank) ]
                str_indexes = ','.join(indexes)
                #tostr_indexes = ','.join([ 'kgen_tostr(idx)' for idx in indexes])

                prevobj = pobj
                doobjs = []
                for d in range(var.rank):
                    attrs = {'loopcontrol': 'idx%(d)d=kgen_bound(1,%(d)d), kgen_bound(2,%(d)d)'%{'d':d+1}}
                    doobj = part_append_genknode(prevobj, EXEC_PART, block_statements.Do, attrs=attrs)
                    doobjs.append(doobj)
                    prevobj = doobj

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_genknode(doobjs[-1], EXEC_PART, block_statements.IfThen, attrs=attrs)

                callname = None
                for uname, req in stmt.unknowns.items():
                    if ( is_class_derived and uname.firstpartname()==stmt.selector[1] ) or uname.firstpartname()==stmt.name:
                        if len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_readname(res)
                            break
                if callname is None:
                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                    part_append_comment(ifpvarobj, EXEC_PART, \
                        'ERROR: "%s" is not resolved. Call statements to read "%s" is not created here.'%\
                        (stmt.name, stmt.name))
                else:
                    attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, '.TRUE.']}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                    pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                    attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, pstr]}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

            else: # intrinsic type
                attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                if stmt.is_numeric():
                    attrs = {'designator': 'kgen_array_sumcheck', 'items': ['printname', \
                        'kgen_array_sum', 'DBLE(SUM(var, mask=(var .eq. var)))', '.TRUE.']}
                    part_append_genknode(pobj, EXEC_PART, statements.Call, attrs=attrs)

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
                    
                if stmt.is_numeric():
                    attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // ")) = "', 'DBLE(SUM(var, mask=(var .eq. var)))']}
                else:   
                    attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

        else: # scalar

            if var.is_allocatable() or var.is_pointer() or allocate:
                attrs = {'items': ['var']}
                part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

            if stmt.is_derived() or is_class_derived:
                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                callname = None
                for uname, req in stmt.unknowns.items():
                    if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                        if len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_readname(res)
                            break
                if callname is None:
                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                    part_append_comment(ifpvarobj, EXEC_PART, \
                        'ERROR: "%s" is not resolved. Call statements to read "%s" is not created here.'%\
                        (stmt.name, stmt.name))
                else:
                    attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', '.TRUE.']}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                    pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                    attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', pstr]}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

            else: # intrinsic type
                attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

def create_write_subr(subrname, entity_name, parent, var, stmt, implicit=False):
    checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname

    is_class_derived = check_class_derived(stmt)

    if not part_has_node(parent, SUBP_PART, checks):

        checks = lambda n: n.kgen_match_class==statements.Contains
        if not parent in state_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
            part_append_comment(parent, CONTAINS_PART, '')
            part_append_gensnode(parent, CONTAINS_PART, statements.Contains)
            part_append_comment(parent, CONTAINS_PART, '')
            state_gencore_contains.append(parent)

        part_append_comment(parent, SUBP_PART, 'write state subroutine for %s'%subrname)
        attrs = {'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
        subrobj = part_append_gensnode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(parent, SUBP_PART, '')

        parent.kgen_stmt.top.used4genstate = True

        # variable A
        #import pdb; pdb.set_trace()
        attrspec = ['INTENT(IN)']
        if var.is_pointer(): attrspec.append('POINTER')
        if var.is_allocatable(): attrspec.append('ALLOCATABLE')
        if var.is_array(): attrspec.append('DIMENSION(%s)'% ','.join(':'*var.rank))
        attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, 'selector':stmt.selector, 'entity_decls': ['var']}
        part_append_gensnode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

        # kgen_unit
        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
        part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        # printname
        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
        part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # printvar
        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
        part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # kgen_istrue
        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
        part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
        part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

        # array index A
        if var.is_array():
            attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'idx%d'%(d+1) for d in range(var.rank) ]}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        part_append_comment(subrobj, DECL_PART, '')

        pobj = gen_write_istrue(subrobj, var, 'var')
        part_append_comment(subrobj, EXEC_PART, '')

        if var.is_array():
            for dim in range(var.rank):
                attrs = {'items': ['LBOUND(var, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                attrs = {'items': ['UBOUND(var, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

            if stmt.is_derived() or is_class_derived:
                indexes = [ 'idx%d'%(d+1) for d in range(var.rank) ]
                str_indexes = ','.join(indexes)
                #tostr_indexes = ','.join([ 'kgen_tostr(idx)' for idx in indexes])

                prevobj = pobj
                doobjs = []
                for d in range(var.rank):
                    attrs = {'loopcontrol': 'idx%(d)d=LBOUND(var,%(d)d), UBOUND(var,%(d)d)'%{'d':d+1}}
                    doobj = part_append_gensnode(prevobj, EXEC_PART, block_statements.Do, attrs=attrs)

                    doobjs.append(doobj)
                    prevobj = doobj

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_gensnode(doobjs[-1], EXEC_PART, block_statements.IfThen, attrs=attrs)

                callname = None
                for uname, req in stmt.unknowns.items():
                    if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                    #if uname.firstpartname()==stmt.name:
                        if len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_writename(res)
                            break
                if callname is None:
                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                    part_append_comment(ifpvarobj, EXEC_PART, \
                        'ERROR: "%s" is not resolved. Call statements to write "%s" is not created here.'%\
                        (stmt.name, stmt.name))
                else:
                    attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, '.TRUE.']}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                    pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                    attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, pstr]}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

            else: # intrinsic type
                attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

        else: # scalar
            if stmt.is_derived() or is_class_derived:

                callname = None
                for uname, req in stmt.unknowns.items():
                    if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                    #if uname.firstpartname()==stmt.name:
                        if len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_writename(res)
                            break
                if callname is None:
                    print('WARNING: Can not find Type resolver for %s'%stmt.name)
                    part_append_comment(pobj, EXEC_PART, \
                        'ERROR: "%s" is not resolved. Call statements to write "%s" is not created here.'%\
                        (stmt.name, stmt.name))
                else:
                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', '.TRUE.']}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                    pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                    attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', pstr]}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

            else: # intrinsic type
                attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)




# gen_write_typedecl_in_module.py
from __future__ import absolute_import
 
from collections import OrderedDict
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

from .gencore_utils import get_topname, get_typedecl_writename, get_dtype_writename, get_module_in_writename, STATE_PBLOCK_WRITE_IN_EXTERNS, \
    STATE_PBLOCK_USE_PART, kernel_gencore_contains, state_gencore_contains, get_typedecl_readname, get_dtype_readname, get_module_in_readname, \
    KERNEL_PBLOCK_USE_PART, DRIVER_READ_IN_EXTERNS, process_spec_stmts, get_module_out_writename, get_module_out_readname, \
    KERNEL_PBLOCK_READ_OUT_EXTERNS, STATE_PBLOCK_WRITE_OUT_EXTERNS, gen_write_istrue, gen_read_istrue, is_excluded, \
    is_remove_state, is_zero_array, DRIVER_USE_PART, check_class_derived
from .gencore_subr import create_write_subr, create_read_subr

class Gen_Typedecl_In_Module(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

        self.state_externs_subrs = OrderedDict()
        self.kernel_externs_subrs = OrderedDict()

        self.state_callsite_use_stmts = []
        self.kernel_callsite_use_stmts = []

        self.state_callsite_call_stmts = []
        self.kernel_callsite_call_stmts = []

        self.state_created_subrs = []
        self.kernel_created_subrs = []

        self.state_extern_writes = []
        self.kernel_extern_reads = []

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Module, self.has_externs_in_module, self.create_state_module_parts) 

        #self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
        #    block_statements.Module, None, self.use_ieee_module) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Module, self.has_externs_in_module, self.create_kernel_module_parts) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Module, None, self.add_default_stmts) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Module, self.has_specstmts_in_module, self.process_specstmts_in_module) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Use, None, self.process_use_in_module) 

    def has_externs_in_module(self, node):
        for stmt in node.kgen_stmt.content:
            if isinstance(stmt, typedecl_statements.TypeDeclarationStatement) and \
                "parameter" not in stmt.attrspec and  hasattr(stmt, 'geninfo') and \
                any(len(v) > 0 for v in stmt.geninfo.values()):
                for entity_name in [ get_entity_name(decl) for decl in stmt.entity_decls ]:
                    var = stmt.get_variable(entity_name)
                    if not var.is_parameter():
                        return True
        return False

    def has_specstmts_in_module(self, node):
        if not node.kgen_stmt: return False
        if hasattr(node.kgen_stmt, 'spec_stmts'): return True
        else: return False

    def is_extern_in_kernel_module(self, node):
        if node.kgen_stmt and hasattr(node.kgen_stmt, 'geninfo') and any(len(v) > 0 for v in node.kgen_stmt.geninfo.values()) and \
            isinstance(node.kgen_parent.kgen_stmt, block_statements.Module) and 'parameter' not in node.kgen_stmt.attrspec:
            for entity_name in [ get_entity_name(decl) for decl in node.kgen_stmt.entity_decls ]:
                var = node.kgen_stmt.get_variable(entity_name)
                if not var.is_parameter():
                    return True
        return False

    def is_extern_in_state_module(self, node):
        if node.kgen_stmt and hasattr(node.kgen_stmt, 'geninfo') and any(len(v) > 0 for v in node.kgen_stmt.geninfo.values()) and \
            isinstance(node.kgen_parent.kgen_stmt, block_statements.Module) and 'parameter' not in node.kgen_stmt.attrspec:
            for entity_name in [ get_entity_name(decl) for decl in node.kgen_stmt.entity_decls ]:
                var = node.kgen_stmt.get_variable(entity_name)
                if not var.is_parameter():
                    return True
        return False

    def process_specstmts_in_module(self, node):
        process_spec_stmts(node.kgen_stmt)

    def process_use_in_module(self, node):
        if not node.kgen_isvalid: return
        if not node.kgen_stmt: return
        if not hasattr(node.kgen_stmt, 'geninfo'):
            node.kgen_isvalid = False
            return
      
        new_items = []
        unames = list(set([ uname.firstpartname() for uname, req in KGGenType.get_state(node.kgen_stmt.geninfo) ]))
        for item in node.kgen_stmt.items:
            if item.split('=>')[0].strip() in unames:
                new_items.append(item)
        node.items = new_items
        node.nature = node.kgen_stmt.nature
        node.isonly = node.kgen_stmt.isonly
        node.kgen_use_tokgen = True

#    def use_ieee_module(self, node):
#
#        attrs = {'name':'IEEE_ARITHMETIC', 'nature': 'INTRINSIC', 'isonly': True, 'items':['ieee_is_normal']}
#        part_append_gensnode(node, USE_PART, statements.Use, attrs=attrs)


    def add_default_stmts(self, node):

        attrs = {'name':'kgen_utils_mod', 'isonly': True, 'items':['kgen_dp', 'kgen_array_sumcheck']}
        part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        #attrs = {'name':'tprof_mod', 'isonly': True, 'items':['tstart', 'tstop', 'tnull', 'tprnt']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        #attrs = {'name':'IEEE_ARITHMETIC', 'nature': 'INTRINSIC', 'isonly': True, 'items':['ieee_is_normal']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

    def create_kernel_module_parts(self, node):
        in_subrobj = None
        in_subrname = get_module_in_readname(node.kgen_stmt)
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==in_subrname
        if not part_has_node(node, SUBP_PART, checks):

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
            if not node in kernel_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
                part_append_comment(node, CONTAINS_PART, '')
                part_append_genknode(node, CONTAINS_PART, statements.Contains)
                part_append_comment(node, CONTAINS_PART, '')
                kernel_gencore_contains.append(node)

            attrs = {'name': in_subrname, 'args': ['kgen_unit']}
            part_append_comment(node, SUBP_PART, 'read state subroutine for %s'%in_subrname)
            in_subrobj = part_append_genknode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(node, SUBP_PART, '')

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_genknode(in_subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_genknode(in_subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_genknode(in_subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            part_append_comment(in_subrobj, DECL_PART, '')

            # add public stmt
            attrs = {'items':[in_subrname]}
            part_append_genknode(node, DECL_PART, statements.Public, attrs=attrs)

        out_subrobj = None
        if hasattr(node.kgen_stmt, 'geninfo') and KGGenType.has_state_out(node.kgen_stmt.geninfo):
            out_subrname = get_module_out_readname(node.kgen_stmt)
            checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==out_subrname
            if not part_has_node(node, SUBP_PART, checks):

                checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
                if not node in kernel_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
                    part_append_comment(node, CONTAINS_PART, '')
                    part_append_genknode(node, CONTAINS_PART, statements.Contains)
                    part_append_comment(node, CONTAINS_PART, '')
                    kernel_gencore_contains.append(node)

                attrs = {'name': out_subrname, 'args': ['kgen_unit']}
                part_append_comment(node, SUBP_PART, 'read state subroutine for %s'%out_subrname)
                out_subrobj = part_append_genknode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
                part_append_comment(node, SUBP_PART, '')

                # kgen_unit
                attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
                part_append_genknode(out_subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)
                part_append_comment(out_subrobj, DECL_PART, '')

                # kgen_istrue
                attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
                part_append_genknode(out_subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

                attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
                part_append_genknode(out_subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

                # add public stmt
                attrs = {'items':[out_subrname]}
                part_append_genknode(node, DECL_PART, statements.Public, attrs=attrs)


        if in_subrobj or out_subrobj:
            self.kernel_externs_subrs[node] = ( in_subrobj, out_subrobj )

            # register event per typedecl 
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
                typedecl_statements.TypeDeclarationStatement, self.is_extern_in_kernel_module, self.create_subr_read_typedecl_in_module) 

            # register event per module
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
                block_statements.Module, self.has_externs_in_module, self.create_kernel_stmts_in_callsite) 

    def create_state_module_parts(self, node):

        in_subrname = get_module_in_writename(node.kgen_stmt)
        in_subrobj = None
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==in_subrname
        if not part_has_node(node, SUBP_PART, checks):

            checks = lambda n: n.kgen_match_class==statements.Contains
            if not node in state_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
                part_append_comment(node, CONTAINS_PART, '')
                part_append_gensnode(node, CONTAINS_PART, statements.Contains)
                part_append_comment(node, CONTAINS_PART, '')
                state_gencore_contains.append(node)

            attrs = {'name': in_subrname, 'args': ['kgen_unit']}
            part_append_comment(node, SUBP_PART, 'write in state subroutine for %s'%in_subrname)
            in_subrobj = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(node, SUBP_PART, '')

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_gensnode(in_subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_gensnode(in_subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_gensnode(in_subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            part_append_comment(in_subrobj, DECL_PART, '')

            # add public stmt
            attrs = {'items':[in_subrname]}
            part_append_gensnode(node, DECL_PART, statements.Public, attrs=attrs)

        out_subrobj = None
        if hasattr(node.kgen_stmt, 'geninfo') and KGGenType.has_state_out(node.kgen_stmt.geninfo):
            out_subrname = get_module_out_writename(node.kgen_stmt)
            checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==out_subrname
            if not part_has_node(node, SUBP_PART, checks):

                checks = lambda n: n.kgen_match_class==statements.Contains
                if not node in state_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
                    part_append_comment(node, CONTAINS_PART, '')
                    part_append_gensnode(node, CONTAINS_PART, statements.Contains)
                    part_append_comment(node, CONTAINS_PART, '')
                    state_gencore_contains.append(node)

                attrs = {'name': out_subrname, 'args': ['kgen_unit']}
                part_append_comment(node, SUBP_PART, 'write out state subroutine for %s'%out_subrname)
                out_subrobj = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
                part_append_comment(node, SUBP_PART, '')

                # kgen_unit
                attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
                part_append_gensnode(out_subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                # kgen_istrue
                attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
                part_append_gensnode(out_subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

                attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
                part_append_gensnode(out_subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

                part_append_comment(out_subrobj, DECL_PART, '')

                # add public stmt
                attrs = {'items':[out_subrname]}
                part_append_gensnode(node, DECL_PART, statements.Public, attrs=attrs)

        if in_subrobj or out_subrobj:
            self.state_externs_subrs[node] = (in_subrobj, out_subrobj)

            node.kgen_stmt.top.used4genstate = True

            # register event per typedecl 
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
                typedecl_statements.TypeDeclarationStatement, self.is_extern_in_state_module, self.create_subr_write_typedecl_in_module) 

            # register event per module
            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
                block_statements.Module, self.has_externs_in_module, self.create_state_stmts_in_callsite) 
        else:
            raise Exception('Dupulicated state extern subroutine name for module: %s. Please ensure that KGen-generated source file is NOT re-used.'%node.name)

    def create_kernel_stmts_in_callsite(self, node):
        if not self.kernel_externs_subrs[node][0] in self.kernel_callsite_use_stmts:
            attrs = {'name':node.name, 'isonly': True, 'items':[self.kernel_externs_subrs[node][0].name]}
            namedpart_append_genknode(node.kgen_kernel_id, DRIVER_USE_PART, statements.Use, attrs=attrs)
            self.kernel_callsite_use_stmts.append(self.kernel_externs_subrs[node][0])

        if not self.kernel_externs_subrs[node][0] in self.kernel_callsite_call_stmts:
            attrs = {'designator': self.kernel_externs_subrs[node][0].name, 'items': ['kgen_unit']}
            namedpart_append_genknode(node.kgen_kernel_id, DRIVER_READ_IN_EXTERNS, statements.Call, attrs=attrs)
            self.kernel_callsite_call_stmts.append(self.kernel_externs_subrs[node][0])

        if hasattr(node.kgen_stmt, 'geninfo') and KGGenType.has_state_out(node.kgen_stmt.geninfo):
            if not self.kernel_externs_subrs[node][1] in self.kernel_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
                attrs = {'name':node.name, 'isonly': True, 'items':[self.kernel_externs_subrs[node][1].name]}
                namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_USE_PART, statements.Use, attrs=attrs)
                self.kernel_callsite_use_stmts.append(self.kernel_externs_subrs[node][1])

            if not self.kernel_externs_subrs[node][1] in self.kernel_callsite_call_stmts:
                attrs = {'designator': self.kernel_externs_subrs[node][1].name, 'items': ['kgen_unit']}
                namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, statements.Call, attrs=attrs)
                self.kernel_callsite_call_stmts.append(self.kernel_externs_subrs[node][1])

#
#        if not self.kernel_externs_subrs[node][0] in self.kernel_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
#            attrs = {'name':node.name, 'isonly': True, 'items':[self.kernel_externs_subrs[node][0].name]}
#            namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_USE_PART, statements.Use, attrs=attrs)
#            self.kernel_callsite_use_stmts.append(self.kernel_externs_subrs[node][0])
#
#        if not self.kernel_externs_subrs[node][0] in self.kernel_callsite_call_stmts:
#            attrs = {'designator': self.kernel_externs_subrs[node][0].name, 'items': ['kgen_unit']}
#            namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_READ_IN_EXTERNS, statements.Call, attrs=attrs)
#            self.kernel_callsite_call_stmts.append(self.kernel_externs_subrs[node][0])
#
#        if hasattr(node.kgen_stmt, 'geninfo') and KGGenType.has_state_out(node.kgen_stmt.geninfo):
#            if not self.kernel_externs_subrs[node][1] in self.kernel_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
#                attrs = {'name':node.name, 'isonly': True, 'items':[self.kernel_externs_subrs[node][1].name]}
#                namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_USE_PART, statements.Use, attrs=attrs)
#                self.kernel_callsite_use_stmts.append(self.kernel_externs_subrs[node][1])
#
#            if not self.kernel_externs_subrs[node][1] in self.kernel_callsite_call_stmts:
#                attrs = {'designator': self.kernel_externs_subrs[node][1].name, 'items': ['kgen_unit']}
#                namedpart_append_genknode(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, statements.Call, attrs=attrs)
#                self.kernel_callsite_call_stmts.append(self.kernel_externs_subrs[node][1])

    def create_state_stmts_in_callsite(self, node):
        kgenunit = 'kgen_unit'

        if not self.state_externs_subrs[node][0] in self.state_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
            attrs = {'name':node.name, 'isonly': True, 'items':[self.state_externs_subrs[node][0].name]}
            namedpart_append_gensnode(node.kgen_kernel_id, STATE_PBLOCK_USE_PART, statements.Use, attrs=attrs)
            self.state_callsite_use_stmts.append(self.state_externs_subrs[node][0])

        if not self.state_externs_subrs[node][0] in self.state_callsite_call_stmts:
            attrs = {'designator': self.state_externs_subrs[node][0].name, 'items': [kgenunit]}
            namedpart_append_gensnode(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_EXTERNS, statements.Call, attrs=attrs)
            self.state_callsite_call_stmts.append(self.state_externs_subrs[node][0])

        if hasattr(node.kgen_stmt, 'geninfo') and KGGenType.has_state_out(node.kgen_stmt.geninfo):
            if not self.state_externs_subrs[node][1] in self.state_callsite_use_stmts and node.name!=getinfo('topblock_stmt').name:
                attrs = {'name':node.name, 'isonly': True, 'items':[self.state_externs_subrs[node][1].name]}
                namedpart_append_gensnode(node.kgen_kernel_id, STATE_PBLOCK_USE_PART, statements.Use, attrs=attrs)
                self.state_callsite_use_stmts.append(self.state_externs_subrs[node][1])

            if not self.state_externs_subrs[node][1] in self.state_callsite_call_stmts:
                attrs = {'designator': self.state_externs_subrs[node][1].name, 'items': [kgenunit]}
                namedpart_append_gensnode(node.kgen_kernel_id, STATE_PBLOCK_WRITE_OUT_EXTERNS, statements.Call, attrs=attrs)
                self.state_callsite_call_stmts.append(self.state_externs_subrs[node][1])

    def create_subr_read_typedecl_in_module(self, node):

        parent = node.kgen_parent
        stmt = node.kgen_stmt
        raw_entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state(stmt.geninfo)])
        entity_names = [ e for e in raw_entity_names if not stmt.get_variable(e).is_parameter() ]
        
        raw_out_entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state_out(stmt.geninfo)])
        out_entity_names = [ e for e in raw_out_entity_names if not stmt.get_variable(e).is_parameter() ]

        def get_attrs(attrspec, allowed_attrs):
            attrspec = []
            for attr in stmt.attrspec:
                if any( attr.startswith(allowed_attr) for allowed_attr in allowed_attrs):
                    attrspec.append(attr)
            return attrspec

        def get_decls(names, decls, prefix=''):
            import re
            entity_decls = []
            for decl in decls:
                ename = re.split(r'\(|\*|=', decl)[0].strip()
                if ename in names:
                    entity_decls.append(prefix+decl)
            return entity_decls

        if len(entity_names)==0:
            node.kgen_forced_line = False
        elif len(entity_names)!=len(stmt.entity_decls):
            attrspec = get_attrs(stmt.attrspec, ['pointer', 'allocatable', 'dimension', 'public', 'target'])
            entity_decls = get_decls(entity_names, stmt.entity_decls)

            attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, \
                'selector':stmt.selector, 'entity_decls': entity_decls}

            if stmt.is_derived():
                node.type_spec = 'TYPE'
            else:
                node.type_spec = stmt.__class__.__name__.upper()
            node.attrspec = attrspec
            node.selector = stmt.selector
            node.entity_decls = entity_decls
            node.kgen_use_tokgen = True
            #part_append_genknode(node.kgen_parent, DECL_PART, stmt.__class__, attrs=attrs)
            #part_append_genknode(node.kgen_parent, DECL_PART, stmt.__class__, attrs=attrs)

        if len(out_entity_names)>0:
            attrspec = get_attrs(stmt.attrspec, ['pointer', 'allocatable', 'dimension'])

            entity_decls = get_decls(out_entity_names, stmt.entity_decls, prefix='kgenref_')

            attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, \
                'selector':stmt.selector, 'entity_decls': entity_decls}
            part_append_genknode(node.kgen_parent, DECL_PART, stmt.__class__, attrs=attrs)

        is_class_derived = check_class_derived(stmt)

        for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
            if node.kgen_parent.name+entity_name in self.kernel_extern_reads: continue

            if is_remove_state(entity_name, stmt): continue

            self.kernel_extern_reads.append(node.kgen_parent.name+entity_name)

            var = stmt.get_variable(entity_name)
            subrname = get_typedecl_readname(stmt, entity_name)

            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                    if entity_name in out_entity_names:
                        self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var, prefix='kgenref_')
                    if subrname not in self.kernel_created_subrs:
                        create_read_subr(subrname, entity_name, parent, var, stmt)
                        self.kernel_created_subrs.append(subrname)
                else: # intrinsic type
                    if var.is_explicit_shape_array():
                        self.create_read_intrinsic(self.kernel_externs_subrs[node.kgen_parent][0], entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_read_intrinsic(self.kernel_externs_subrs[node.kgen_parent][1], entity_name, stmt, var, prefix='kgenref_')
                    else: # implicit array
                        self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var, prefix='kgenref_')
                        if subrname not in self.kernel_created_subrs:
                            create_read_subr(subrname, entity_name, parent, var, stmt)
                            self.kernel_created_subrs.append(subrname)
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    if var.is_allocatable() or var.is_pointer():
                        self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var, prefix='kgenref_')
                        if subrname not in self.kernel_created_subrs:
                            create_read_subr(subrname, entity_name, parent, var, stmt)
                            self.kernel_created_subrs.append(subrname)
                    else:
                        subrname = None
                        for uname, req in stmt.unknowns.items():
                            if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                            #if uname.firstpartname()==stmt.name:
                                if len(req.res_stmts)>0:
                                    res = req.res_stmts[0]
                                    subrname = get_dtype_readname(res)
                                    break
                        if subrname is None:
                            print('WARNING: Can not find Type resolver for %s'%stmt.name)
                            namedpart_append_comment(self.kernel_externs_subrs[node.kgen_parent][0], EXEC_PART, \
                                'ERROR: "%s" is not resolved. Call statements to read "%s" is not created here.'%\
                                (stmt.name, stmt.name))
                        else:
                            self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                            if entity_name in out_entity_names:
                                self.create_read_call(self.kernel_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var, prefix='kgenref_')
                else: # intrinsic type
                    self.create_read_intrinsic(self.kernel_externs_subrs[node.kgen_parent][0], entity_name, stmt, var)
                    if entity_name in out_entity_names:
                        self.create_read_intrinsic(self.kernel_externs_subrs[node.kgen_parent][1], entity_name, stmt, var, prefix='kgenref_')

    def create_subr_write_typedecl_in_module(self, node):
        parent = node.kgen_parent
        stmt = node.kgen_stmt

        raw_entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state(stmt.geninfo)])
        entity_names = [ e for e in raw_entity_names if not stmt.get_variable(e).is_parameter() ]
        
        raw_out_entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state_out(stmt.geninfo)])
        out_entity_names = [ e for e in raw_out_entity_names if not stmt.get_variable(e).is_parameter() ]

        #entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state(stmt.geninfo)])
        #out_entity_names = set([ uname.firstpartname() for uname, req in KGGenType.get_state_out(stmt.geninfo)])

        is_class_derived = check_class_derived(stmt)

        for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
            if node.kgen_parent.name+entity_name in self.state_extern_writes: continue
            if is_remove_state(entity_name, stmt): continue

            self.state_extern_writes.append(node.kgen_parent.name+entity_name)

            var = stmt.get_variable(entity_name)
            subrname = get_typedecl_writename(stmt, entity_name)

            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    self.create_write_call(self.state_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                    if entity_name in out_entity_names:
                        self.create_write_call(self.state_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var)
                    if subrname not in self.state_created_subrs:
                        create_write_subr(subrname, entity_name, parent, var, stmt)
                        self.state_created_subrs.append(subrname)
                else: # intrinsic type
                    if var.is_explicit_shape_array():
                        self.create_write_intrinsic(self.state_externs_subrs[node.kgen_parent][0], entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_write_intrinsic(self.state_externs_subrs[node.kgen_parent][1], entity_name, stmt, var)
                    else: # implicit array
                        self.create_write_call(self.state_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_write_call(self.state_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var)
                        if subrname not in self.state_created_subrs:
                            create_write_subr(subrname, entity_name, parent, var, stmt)
                            self.state_created_subrs.append(subrname)
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    if var.is_allocatable() or var.is_pointer():
                        self.create_write_call(self.state_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                        if entity_name in out_entity_names:
                            self.create_write_call(self.state_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var)
                        if subrname not in self.state_created_subrs:
                            create_write_subr(subrname, entity_name, parent, var, stmt)
                            self.state_created_subrs.append(subrname)
                    else:
                        subrname = None
                        for uname, req in stmt.unknowns.items():
                            if ( is_class_derived and uname.firstpartname()==stmt.selector[1]) or uname.firstpartname()==stmt.name:
                            #if uname.firstpartname()==stmt.name:
                                if len(req.res_stmts)>0:
                                    res = req.res_stmts[0]
                                    subrname = get_dtype_writename(res)
                                    break
                        if subrname is None:
                            print('WARNING: Can not find Type resolver for %s'%stmt.name)
                            namedpart_append_comment(self.state_externs_subrs[node.kgen_parent][0], EXEC_PART, \
                                'ERROR: "%s" is not resolved. Call statements to write "%s" is not created here.'%\
                                (stmt.name, stmt.name))
                        else:
                            self.create_write_call(self.state_externs_subrs[node.kgen_parent][0], subrname, entity_name, stmt, var)
                            if entity_name in out_entity_names:
                                self.create_write_call(self.state_externs_subrs[node.kgen_parent][1], subrname, entity_name, stmt, var)
                else: # intrinsic type
                    self.create_write_intrinsic(self.state_externs_subrs[node.kgen_parent][0], entity_name, stmt, var)
                    if entity_name in out_entity_names:
                        self.create_write_intrinsic(self.state_externs_subrs[node.kgen_parent][1], entity_name, stmt, var)

    def create_read_intrinsic(self, subrobj, entity_name, stmt, var, prefix=''):

        pobj = gen_read_istrue(subrobj, var, entity_name)

        if var.is_allocatable() or var.is_pointer():
            attrs = {'items': ['%s'%(prefix+entity_name)]}
            part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': [prefix+entity_name], 'specs': ['UNIT = kgen_unit']}
        part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

        if var.is_array() and stmt.is_numeric():
            attrs = {'designator': 'kgen_array_sumcheck', 'items': ['"%s"'%(prefix+entity_name), \
                'kgen_array_sum', 'DBLE(SUM(%s, mask=(%s .eq. %s)))'%(prefix+entity_name, prefix+entity_name, prefix+entity_name), '.TRUE.']}
            part_append_genknode(pobj, EXEC_PART, statements.Call, attrs=attrs)

        if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):
            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(%s)) = "'%(prefix+entity_name), 'DBLE(SUM(%s, mask=(%s .eq. %s)))'%(prefix+entity_name, prefix+entity_name, prefix+entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: %s = "'%(prefix+entity_name), prefix+entity_name]}
            part_append_genknode(pobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_write_intrinsic(self, subrobj, entity_name, stmt, var):

        pobj = gen_write_istrue(subrobj, var, entity_name)

        attrs = {'items': [entity_name], 'specs': ['UNIT = kgen_unit']}
        part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

        if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):
            if stmt.is_numeric() and var.is_array():
                attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(%s)) = "'%(prefix+entity_name), 'DBLE(SUM(%s, mask=(%s .eq. %s)))'%(entity_name, entity_name, entity_name)]}
            else:
                attrs = {'items': ['"KGEN DEBUG: %s = "'%(prefix+entity_name), entity_name]}
            part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_read_call(self, subrobj, callname, entity_name, stmt, var, prefix=''):

        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
        attrs = {'designator': callname, 'items': [prefix+entity_name, 'kgen_unit', '"%s"'%(prefix+entity_name), pstr]}
        part_append_genknode(subrobj, EXEC_PART, statements.Call, attrs=attrs)

    def create_write_call(self, subrobj, callname, entity_name, stmt, var, prefix=''):

        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
        attrs = {'designator': callname, 'items': [entity_name, 'kgen_unit', '"%s"'%(prefix+entity_name), pstr]}
        part_append_gensnode(subrobj, EXEC_PART, statements.Call, attrs=attrs)




# gen_read_callsite_file.py
from __future__ import absolute_import

from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

# TODO: adjust in and out state of variables for verification

from .gencore_utils import (KERNEL_PBLOCK_USE_PART, KERNEL_PBLOCK_DECL_PART, KERNEL_PBLOCK_EXEC_PART,
    KERNEL_PBLOCK_CONTAINS_PART, KERNEL_PBLOCK_SUBP_PART, KERNEL_PBLOCK_READ_IN_LOCALS,
    KERNEL_PBLOCK_READ_OUT_EXTERNS, KERNEL_PBLOCK_READ_OUT_LOCALS, KERNEL_TBLOCK_USE_PART, KERNEL_TBLOCK_DECL_PART,
    KERNEL_TBLOCK_CONTAINS_PART, KERNEL_TBLOCK_SUBP_PART, kernel_gencore_contains, KERNEL_PBLOCK_BEFORE_KERNEL, KERNEL_PBLOCK_AFTER_KERNEL,
    KERNEL_PBLOCK_PREPROCESS)

class Gen_K_Callsite_File(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('callsite_stmts')[0], None, self.create_callsite_parts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            getinfo('callsite_stmts')[0], None, self.invalid_kernel_stmts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('parentblock_stmt'), None, self.create_parentblock_parts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            getinfo('parentblock_stmt'), None, self.set_args)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('topblock_stmt'), None, self.create_topblock_parts)

    def set_args(self, node):

        callsite_args = getinfo('kernel_driver_callsite_args')[:]

        node.new_args = callsite_args
        node.tosubr = True
        node.kgen_use_tokgen = True

        node.kgen_end_obj.tosubr = True
        node.kgen_end_obj.kgen_use_tokgen = True

    def create_parentblock_parts(self, node):

        namedpart_link_part(node, KERNEL_PBLOCK_USE_PART, USE_PART)
        namedpart_link_part(node, KERNEL_PBLOCK_DECL_PART, DECL_PART)
        namedpart_link_part(node, KERNEL_PBLOCK_EXEC_PART, EXEC_PART)
        namedpart_link_part(node, KERNEL_PBLOCK_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, KERNEL_PBLOCK_SUBP_PART, SUBP_PART)

        attrs = {'name':'kgen_utils_mod', 'isonly': True, 'items':['kgen_dp', 'kgen_array_sumcheck']}
        part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)
           
        #attrs = {'name':'IEEE_ARITHMETIC', 'nature': 'INTRINSIC', 'isonly': True, 'items':['ieee_is_normal']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(node, DECL_PART, '#ifdef KGEN_PAPI', style='rawtext')

            if getinfo('papi_header_file') is not None:
                part_append_comment(node, DECL_PART, '#include <%s>'%getinfo('papi_header_file'), style='rawtext')
                
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(OUT)'], 'entity_decls': ['kgen_measure'], \
                'selector': (None, '8')}
            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            part_append_comment(node, DECL_PART, '#else', style='rawtext')

        attrs = {'type_spec': 'REAL', 'attrspec': ['INTENT(OUT)'], 'entity_decls': ['kgen_measure'], \
            'selector': (None, 'kgen_dp')}
        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(node, DECL_PART, '#endif', style='rawtext')

        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(OUT)'], 'entity_decls': ['kgen_isverified']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'selector': ('*', None), 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_filepath']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_intvar', 'kgen_ierr']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke', 'kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage') ) ]}
        part_append_genknode(node, DECL_PART, statements.Common, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['PARAMETER'], 'entity_decls': ['KGEN_MAXITER = 1']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        if getinfo('cache_pollution'):

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['PARAMETER'], 'entity_decls': ['CACHE_CONTROL_LENGTH=%d'%getinfo('cache_pollution')]}
            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['DIMENSION(CACHE_CONTROL_LENGTH)'], 'entity_decls': ['kgen_cache_control']}
            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        part_append_comment(node, DECL_PART, '')

        namedpart_create_subpart(node, KERNEL_PBLOCK_PREPROCESS, EXEC_PART, index=0)
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '')
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, 'parent block preprocessing')

        if getinfo('add_mpi_frame'):
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '')
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '#ifdef _MPI', style='rawtext')
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, 'call mpi_comm_rank(mpi_comm_world, kgen_mpirank, kgen_ierr)', style='rawtext')
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '#else', style='rawtext')

        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, 'kgen_mpirank = 0', style='rawtext')

        if getinfo('add_mpi_frame'):
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '#endif', style='rawtext')
            namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_PREPROCESS, '', style='rawtext')

        namedpart_create_subpart(node, KERNEL_PBLOCK_READ_IN_LOCALS, EXEC_PART, index=1)
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_IN_LOCALS, '')
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_IN_LOCALS, 'local input variables')

        namedpart_create_subpart(node, KERNEL_PBLOCK_READ_OUT_EXTERNS, EXEC_PART, index=2)
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, '')
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, 'extern output variables')

        namedpart_create_subpart(node, KERNEL_PBLOCK_READ_OUT_LOCALS, EXEC_PART, index=3)
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_LOCALS, '')
        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_LOCALS, 'local output variables')

#        kernel_stmts = getinfo('callsite_stmts')
#        if len(kernel_stmts)!=1 or not isinstance(kernel_stmts[0], statements.Call):
#            # ensure contains
#            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
#            if not node in kernel_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
#                part_append_comment(node, CONTAINS_PART, '')
#                part_append_genknode(node, CONTAINS_PART, statements.Contains)
#                part_append_comment(node, CONTAINS_PART, '')
#                kernel_gencore_contains.append(node)
#
#
#            part_append_comment(node, SUBP_PART, 'kgen kernel subroutine')
#            attrs = {'name': 'kgen_kernel'}
#            subrobj = part_append_genknode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
#            part_append_comment(node, SUBP_PART, '')
#
#            start = kernel_stmts[0].item.span[0]-1
#            end = kernel_stmts[-1].item.span[1]
#            lines = kernel_stmts[0].top.prep[start:end]
#            lines_str = '\n'.join(lines)
#            dummy_node = part_append_genknode(subrobj, EXEC_PART, statements.Call)
#            dummy_node.kgen_stmt = getinfo('dummy_stmt')
#            dummy_node.kgen_forced_line = lines_str
        # debug
#        attrs = {'variable': 'kgen_region_started', 'sign': '=', 'expr': '.TRUE.'}
#        part_insert_gensnode(node, EXEC_PART, statements.Assignment, attrs=attrs, index=0)
#
#        attrs = {'variable': 'kgen_region_started', 'sign': '=', 'expr': '.FALSE.'}
#        part_append_gensnode(node, EXEC_PART, statements.Assignment, attrs=attrs)

    def create_topblock_parts(self, node):

        namedpart_link_part(node, KERNEL_TBLOCK_USE_PART, USE_PART)
        namedpart_link_part(node, KERNEL_TBLOCK_DECL_PART, DECL_PART)
        namedpart_link_part(node, KERNEL_TBLOCK_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, KERNEL_TBLOCK_SUBP_PART, SUBP_PART)

        # ensure public for parentblock stmt
        checks = lambda n: isinstance(n.kgen_stmt, statements.Public) and n.kgen_stmt.items and \
            getinfo('parentblock_stmt').name in n.kgen_stmt.items
        if not part_has_node(node, DECL_PART, checks):
            attrs = {'items':[getinfo('parentblock_stmt').name]}
            part_append_genknode(node, DECL_PART, statements.Public, attrs=attrs)

        if getinfo('add_mpi_frame'):
            part_append_comment(node, DECL_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, DECL_PART, 'include "mpif.h"', style='rawtext')
            part_append_comment(node, DECL_PART, '#endif', style='rawtext')
            part_append_comment(node, DECL_PART, '')

    def create_callsite_parts(self, node):
        index, partname, part = get_part_index(node)

#        namedpart_create_subpart(node.kgen_parent, KERNEL_PBLOCK_READ_IN_LOCALS, EXEC_PART, index=index)
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_IN_LOCALS, '')
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_IN_LOCALS, 'local input variables')
#
#        namedpart_create_subpart(node.kgen_parent, KERNEL_PBLOCK_READ_OUT_EXTERNS, EXEC_PART, index=index+1)
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, '')
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_EXTERNS, 'extern output variables')
#
#        namedpart_create_subpart(node.kgen_parent, KERNEL_PBLOCK_READ_OUT_LOCALS, EXEC_PART, index=index+2)
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_LOCALS, '')
#        namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_READ_OUT_LOCALS, 'local output variables')

        idx = index

        attrs = {'expr': 'kgen_evalstage'}
        ifeval = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_beforecallsite_eval', ifeval)
        idx += 1

        attrs = {'expr': 'kgen_warmupstage'}
        ifwarmup = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_beforecallsite_warmup', ifwarmup)
        idx += 1

        attrs = {'expr': 'kgen_mainstage'}
        ifmain = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_beforecallsite_main', ifmain)
        idx += 1

        namedpart_create_subpart(node.kgen_parent, KERNEL_PBLOCK_BEFORE_KERNEL, EXEC_PART, index=idx)
        self.plugin_common[node.kgen_kernel_id]['ext.gencore']['blocks']['before_kernel'] = KERNEL_PBLOCK_BEFORE_KERNEL
        idx += 1

        part_insert_comment(node.kgen_parent, EXEC_PART, idx, '')
        idx += 1
        part_insert_comment(node.kgen_parent, EXEC_PART, idx, 'call to kgen kernel')
        idx += 1

        #attrs = {'variable': 'kgen_warmupstage', 'sign': '=', 'expr': '.TRUE.'}
        #part_insert_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs, index=idx)
        idx += 1
#
#        kernel_stmts = getinfo('callsite_stmts')
#        if len(kernel_stmts)!=1 or not isinstance(kernel_stmts[0], statements.Call):
#            attrs = {'designator': 'kgen_kernel'}
#            part_insert_genknode(node.kgen_parent, EXEC_PART, statements.Call, attrs=attrs, index=index+3)
#        else:
#            start = node.kgen_stmt.item.span[0]-1
#            end = node.kgen_stmt.item.span[1]
#            lines = node.kgen_stmt.top.prep[start:end]
#            lines_str = '\n'.join(lines)
#            dummy_node = part_insert_genknode(node.kgen_parent, EXEC_PART, statements.Call, index=index+3)
#            dummy_node.kgen_stmt = node.kgen_stmt
#            dummy_node.kgen_forced_line = lines_str

        kernel_stmts = getinfo('callsite_stmts')
        start = kernel_stmts[0].item.span[0]-1
        end = kernel_stmts[-1].item.span[1]
        lines = kernel_stmts[0].top.prep[start:end]
        lines_str = '\n'.join(lines)
        dummy_node = part_insert_genknode(node.kgen_parent, EXEC_PART, statements.Call, index=idx)
        dummy_node.kgen_stmt = getinfo('dummy_stmt')
        dummy_node.kgen_forced_line = lines_str
        idx += 1

        namedpart_create_subpart(node.kgen_parent, KERNEL_PBLOCK_AFTER_KERNEL, EXEC_PART, index=idx)
        self.plugin_common[node.kgen_kernel_id]['ext.gencore']['blocks']['after_kernel'] = KERNEL_PBLOCK_AFTER_KERNEL
        idx += 1

        attrs = {'expr': 'kgen_mainstage'}
        ifmain = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_aftercallsite_main', ifmain)
        idx += 1

        attrs = {'expr': 'kgen_warmupstage'}
        ifwarmup = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_aftercallsite_warmup', ifwarmup)
        idx += 1

        attrs = {'expr': 'kgen_evalstage'}
        ifeval = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        setinfo('blocknode_aftercallsite_eval', ifeval)

    def invalid_kernel_stmts(self, node):
        kernel_stmts = getinfo('callsite_stmts')

        for stmt in kernel_stmts:
            stmt.genkpair.kgen_forced_line = False



# gen_write_typedecl_in_module.py
from __future__ import absolute_import

import os 
import re 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin
from .gencore_utils import is_zero_array, check_class_derived

class Gen_Write_In_Module(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.FINISH_PROCESS, \
            statements.Comment, self.is_write_directive, self.write_state_func) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            statements.Comment, self.is_write_directive, self.read_state_func) 

    def is_write_directive(self, node):
        return hasattr(node.kgen_stmt, 'write_state')

    def get_orgstmt(self, parts, stmt):

        if len(parts) == 0:
            raise Exception('Wrong number of parts.')

        match = re.match(r'\w+[^\w]*', parts[0].strip()) 
        if match:
            varname = match.group(0).lower()
            for unknown, res in stmt.unknowns.items():
                if unknown.firstpartname() == stmt.name and isinstance(res.res_stmts[0], block_statements.Type):
                    res_stmt = res.res_stmts[0]
                    if len(parts) == 1:
                        return res_stmt.a.variables[varname]
                    else:
                        return self.get_orgstmt(parts[1:], res_stmt)
        else:
            raise Exception('No matched statment is found.')

    def add_useonlyname(self, ancs, unk, res_stmts):

        if isinstance(res_stmts[-1], statements.Use):
            usestmt = res_stmts[-1] 
            upstmt = usestmt.parent
            name = unk.firstpartname()

            processed = False

            if usestmt.name in upstmt.use_stmts.keys():
                for ustmt in upstmt.use_stmts[usestmt.name]:
                    if not ustmt.isonly:
                        processed = True
                        break
                    if name in ustmt.norenames:
                        processed = True
                        break
                    if name in [ newname for oldname, newname in ustmt.renames ]: 
                        processed = True
                        break
            if not processed:
                attrs = {'name': usestmt.name, 'isonly': True, 'items':[name]}
                part_append_genknode(upstmt.genkpair, USE_PART, statements.Use, attrs=attrs)
        else:
            raise Exception('TODO: support for importing selector')

    def write_state_func(self, node):
        index, partname, part = get_part_index(node)
        tstmt = node.kgen_stmt.ancestors()[0]
        pstmt = node.kgen_stmt.ancestors()[-1]
        pnode = pstmt.genspair
        node.kgen_stmt.top.used4genstate = True
        filename = os.path.splitext(os.path.basename(node.kgen_stmt.reader.id))[0]
        lineno = node.kgen_stmt.item.span[0]
        if not hasattr(pnode, '__write_commonpart_statewrite'):
            #if hasattr(tstmt, 'name') and tstmt.name != getinfo('topblock_stmt').name:
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ierr']}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank']}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            if getinfo('is_openmp_app'):
                attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:1023)' ], 'entity_decls': ['kgen_openmp_issave']}
                part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                attrs = {'type_spec': 'INTEGER', 'entity_decls': ['OMP_GET_THREAD_NUM']}
                part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            else:
                attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_openmp_issave']}
                part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmp_issave') ) ]}
            part_append_gensnode(pnode, DECL_PART, statements.Common, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_write_unit']}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            part_append_comment(pnode, DECL_PART, '')

            setattr(pnode, '__write_commonpart_statewrite', True)

        if getinfo('is_openmp_app'):
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE', 'DIMENSION(0:1023)' ], 'entity_decls': ['kgen_previnvoke_L%d = -1'%lineno]}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE', 'DIMENSION(0:1023)'], 'entity_decls': ['kgen_writesubp_invoke_L%d = 0'%lineno]}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)
        else:
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE'], 'entity_decls': ['kgen_writesubp_invoke_L%d = 0'%lineno]}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE'], 'entity_decls': ['kgen_previnvoke_L%d = -1'%lineno]}
            part_append_gensnode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'entity_decls': ['kgen_write_filepath_L%d'%lineno], 'selector':('1024', None)}
        part_append_gensnode(pnode, DECL_PART, typedecl_statements.Character, attrs=attrs)

        idx = index + 1

        if getinfo('is_openmp_app'):

            attrs = {'expr': 'kgen_openmp_issave(OMP_GET_THREAD_NUM()) .GE. 0'}
            ifsave = part_insert_gensnode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)

            idx += 1

            attrs = {'expr': 'kgen_previnvoke_L%d(OMP_GET_THREAD_NUM()) .LT. 0 .OR. kgen_previnvoke_L%d(OMP_GET_THREAD_NUM()) .NE. kgen_openmp_issave(OMP_GET_THREAD_NUM())'%(lineno, lineno)}
            ifprev = part_append_gensnode(ifsave, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_previnvoke_L%d(OMP_GET_THREAD_NUM())'%lineno, 'sign': '=', 'expr': 'kgen_openmp_issave(OMP_GET_THREAD_NUM())'}
            part_append_gensnode(ifprev, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_writesubp_invoke_L%d(OMP_GET_THREAD_NUM())'%lineno, 'sign': '=', 'expr': '0'}
            part_append_gensnode(ifprev, EXEC_PART, statements.Assignment, attrs=attrs)

            l = [ 'kgen_mpirank', '"."', 'OMP_GET_THREAD_NUM()', '"."', 'kgen_openmp_issave(OMP_GET_THREAD_NUM())', '"."',
                'kgen_writesubp_invoke_L%d(OMP_GET_THREAD_NUM())'%lineno]

            part_append_comment(ifsave, EXEC_PART, 'CRITICAL (%s_L%d)'%(filename, lineno), style='openmp')
        else:

            attrs = {'expr': 'kgen_openmp_issave .GE. 0'}
            ifsave = part_insert_gensnode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)

            idx += 1

            attrs = {'expr': 'kgen_previnvoke_L%d .LT. 0 .OR. kgen_previnvoke_L%d .NE. kgen_openmp_issave'%(lineno, lineno)}
            ifprev = part_append_gensnode(ifsave, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_previnvoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_openmp_issave'}
            part_append_gensnode(ifprev, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_writesubp_invoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
            part_append_gensnode(ifprev, EXEC_PART, statements.Assignment, attrs=attrs)

            l = [ 'kgen_mpirank', '"."', '0', '"."', 'kgen_openmp_issave', '"."', 'kgen_writesubp_invoke_L%d'%lineno]

        # file open
        attrs = {'specs': ['kgen_write_filepath_L%d'%lineno, 'FMT="(A,I0,A,I0,A,I0,A,I0)"' ], 'items': [ '"%s/%s.L%d."'%(getinfo('kernel_path'), filename, lineno) ] + l}
        part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'specs': ['NEWUNIT=kgen_write_unit', 'FILE=kgen_write_filepath_L%d'%lineno, 'STATUS="NEW"', 'ACCESS="STREAM"', \
            'FORM="UNFORMATTED"', 'ACTION="WRITE"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
        part_append_gensnode(ifsave, EXEC_PART, statements.Open, attrs=attrs)


        vars = []
        for varstr in node.kgen_stmt.write_state:
            parts = varstr.lower().split('%')
            if len(parts) > 1:
                # search through use stmts until the leaf stmt
                partstmt = node.kgen_stmt

                match = re.match(r'\w+[^\w]*', parts[0].strip()) 
                if match:
                    varname = match.group(0).lower()
                    for unknown, res in partstmt.unknowns.items():
                        if unknown.firstpartname() == varname and res.res_stmts[0].is_derived():
                            partstmt = res.res_stmts[0]
                            break

                var = self.get_orgstmt(parts[1:], partstmt)
                vars.append((varstr, var))
            else:
                match = re.match(r'\w+[^\w]*', varstr.strip()) 
                if match:
                    varname = match.group(0).lower()
                    for unknown, res in node.kgen_stmt.unknowns.items():
                        if unknown.firstpartname() == varname: 
                            res_stmt = res.res_stmts[0]
                            var = res_stmt.get_variable(varname)
                            if var:
                                vars.append((varstr, var))
                            break

        for varstr, var in vars:
            stmt = var.parent
            is_class_derived = check_class_derived(stmt)
            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    parts = varstr.split('%')
                    parts[-1] = var.name
                    newvarstr = '%'.join(parts)

                    for rank in range(var.rank):
                        attrs = {'items': [ 'LBOUND( %s, %d )'%(newvarstr, rank+1) ], 'specs': ['UNIT = kgen_write_unit']}
                        part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)

                        attrs = {'items': [ 'UBOUND( %s, %d )'%(newvarstr, rank+1) ], 'specs': ['UNIT = kgen_write_unit']}
                        part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'items': [ newvarstr ] , 'specs': ['UNIT = kgen_write_unit']}
                    part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    attrs = {'items': [var.name], 'specs': ['UNIT = kgen_write_unit']}
                    part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)


        # file close
        attrs = {'specs': ['UNIT=kgen_write_unit']}
        part_append_gensnode(ifsave, EXEC_PART, statements.Close, attrs=attrs)

        if getinfo('is_openmp_app'):
            part_append_comment(ifsave, EXEC_PART, 'END CRITICAL (%s_L%d)'%(filename, lineno), style='openmp')

            attrs = {'variable': 'kgen_writesubp_invoke_L%d(OMP_GET_THREAD_NUM())'%lineno, 'sign': '=',
                'expr': 'kgen_writesubp_invoke_L%d(OMP_GET_THREAD_NUM()) + 1'%lineno}
            part_append_gensnode(ifsave, EXEC_PART, statements.Assignment, attrs=attrs)
        else:
            attrs = {'variable': 'kgen_writesubp_invoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_writesubp_invoke_L%d + 1'%lineno}
            part_append_gensnode(ifsave, EXEC_PART, statements.Assignment, attrs=attrs)

        part_insert_comment(node.kgen_parent, EXEC_PART, idx+1, '')

    def read_state_func(self, node):
        index, partname, part = get_part_index(node)
        ancs = node.kgen_stmt.ancestors()
        tstmt = ancs[0]
        pstmt = ancs[-1]
        pnode = pstmt.genkpair
        filename = os.path.splitext(os.path.basename(node.kgen_stmt.reader.id))[0]
        lineno = node.kgen_stmt.item.span[0]

        if not hasattr(pnode, '__write_commonpart_stateread'):

            #if hasattr(tstmt, 'name') and tstmt.name != getinfo('topblock_stmt').name:

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ierr']}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_read_unit = -1']}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke']}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage']}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke', 'kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage') ) ]}
            part_append_genknode(pnode, DECL_PART, statements.Common, attrs=attrs)
            # end if hasattr

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ldim1', 'kgen_udim1', 'kgen_ldim2', 'kgen_udim2', \
                'kgen_ldim3', 'kgen_udim3', 'kgen_ldim4', 'kgen_udim4', 'kgen_ldim5', 'kgen_udim5' ]}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE'], \
                'entity_decls': ['kgen_mindim1 = HUGE(0)', 'kgen_maxdim1 = 0', 'kgen_mindim2 = HUGE(0)', 'kgen_maxdim2 = 0', \
                'kgen_mindim3 = HUGE(0)', 'kgen_maxdim3 = 0', 'kgen_mindim4 = HUGE(0)', 'kgen_maxdim4 = 0', 'kgen_mindim5 = HUGE(0)', 'kgen_maxdim5 = 0' ]}
            part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            setattr(pnode, '__write_commonpart_stateread', True)

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['SAVE'], 'entity_decls': ['kgen_readsubp_evalinvoke_L%d = 0'%lineno, 'kgen_readsubp_warmupinvoke_L%d = 0'%lineno, \
            'kgen_readsubp_maininvoke_L%d = 0'%lineno, 'kgen_readsubp_maxinvoke_L%d = 0'%lineno ]}
        part_append_genknode(pnode, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'entity_decls': ['kgen_read_filepath_L%d'%lineno], 'selector':('128', None)}
        part_append_genknode(pnode, DECL_PART, typedecl_statements.Character, attrs=attrs)

        # collect variables for manual state generation
        vars = []
        for varstr in node.kgen_stmt.write_state:
            parts = varstr.lower().split('%')
            if len(parts) > 1:
                # search through use stmts until the leaf stmt
                partstmt = node.kgen_stmt

                match = re.match(r'\w+[^\w]*', parts[0].strip()) 
                if match:
                    varname = match.group(0).lower()
                    for unknown, res in partstmt.unknowns.items():
                        if unknown.firstpartname() == varname and res.res_stmts[0].is_derived():
                            partstmt = res.res_stmts[0]
                            break

                    var = self.get_orgstmt(parts[1:], partstmt)
                    vars.append((varstr, var))
            else:
                match = re.match(r'\w+[^\w]*', varstr.strip()) 
                if match:
                    varname = match.group(0).lower()
                    for unknown, res in node.kgen_stmt.unknowns.items():
                        if unknown.firstpartname() == varname: 
                            res_stmt = res.res_stmts[0]
                            var = res_stmt.get_variable(varname)
                            if var:
                                vars.append((varstr, var))
                            break

        # create typedecls for array of variable pointer
        for varstr, var in vars:
            stmt = var.parent
            is_class_derived = check_class_derived(stmt)

            type_spec = stmt.name.upper()
            attrspec = list(stmt.attrspec)
            if 'save' not in attrspec:
                attrspec.append('SAVE')
            if 'allocatable' not in attrspec:
                attrspec.append('ALLOCATABLE')
            excludes = []
            for spec in attrspec:
                if spec.startswith('dimension'):
                    excludes.append(spec)
                if spec.startswith('pointer'):
                    excludes.append(spec)
            for exclude in excludes:
                attrspec.remove(exclude)
            selector = tuple(stmt.selector)
            entity_decls = [ 'kgen_arr_%s_L%d'%(var.name, lineno) ]

            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    if var.is_explicit_shape_array():
                        attrspec.append('DIMENSION(%s)'%(', '.join([':']*(var.rank+1))))
                    else: # implicit array
                        attrspec.append('DIMENSION(%s)'%(', '.join([':']*(var.rank+1))))
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    attrspec.append('DIMENSION(:)')

            if hasattr(stmt, 'unknowns'):
                for unk, res in stmt.unknowns.items():
                    res_parent_stmt = res.res_stmts[0].ancestors()[-1]
                    if res_parent_stmt not in ancs:
                        if unk.firstpartname() == selector[0]:
                            self.add_useonlyname(ancs, unk, res.res_stmts)
                        elif unk.firstpartname() == selector[1]:
                            self.add_useonlyname(ancs, unk, res.res_stmts)

            attrs = {'type_spec': type_spec, 'attrspec': attrspec, 'selector': selector, 'entity_decls': entity_decls}
            part_append_genknode(pnode, DECL_PART, stmt.__class__, attrs=attrs)

        idx = index + 1

        # if eval stage
        # - count max number of local invocations
        attrs = {'expr': 'kgen_evalstage .AND. .NOT. kgen_warmupstage .AND. .NOT. kgen_mainstage'}
        ifeval = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        idx += 1

        attrs = {'variable': 'kgen_readsubp_maxinvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifeval, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_warmupinvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifeval, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_maininvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifeval, EXEC_PART, statements.Assignment, attrs=attrs)

        for varstr, var in vars:
            attrs = {'expr': 'ALLOCATED(kgen_arr_%s_L%d)'%(var.name, lineno) }
            ifalloc = part_append_genknode(ifeval, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'items': ['kgen_arr_%s_L%d'%(var.name, lineno)]}
            part_append_genknode(ifalloc, EXEC_PART, statements.Deallocate, attrs=attrs)

        l = [ 'kgen_mpirank', '"."', 'kgen_openmptid', '"."', 'kgen_kernelinvoke', '"."', 'kgen_readsubp_evalinvoke_L%d'%lineno]
        attrs = {'specs': ['kgen_read_filepath_L%d'%lineno, 'FMT="(A,I0,A,I0,A,I0,A,I0)"' ], 'items': [ '"%s.L%d."'%(filename, lineno) ] + l}
        part_append_genknode(ifeval, EXEC_PART, statements.Write, attrs=attrs)

        # file open
        attrs = {'specs': ['NEWUNIT=kgen_read_unit', 'FILE=kgen_read_filepath_L%d'%lineno, 'STATUS="OLD"', 'ACCESS="STREAM"', \
            'FORM="UNFORMATTED"', 'ACTION="READ"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
        part_append_genknode(ifeval, EXEC_PART, statements.Open, attrs=attrs)

        for varstr, var in vars:
            stmt = var.parent
            is_class_derived = check_class_derived(stmt)
            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    parts = varstr.split('%')
                    parts[-1] = var.name
                    newvarstr = '%'.join(parts)

                    bounds = []
                    for rank in range(var.rank):
                        attrs = {'items': [ 'kgen_ldim%d'%(rank+1) ], 'specs': ['UNIT = kgen_read_unit']}
                        part_append_genknode(ifeval, EXEC_PART, statements.Read, attrs=attrs)

                        attrs = {'expr': 'kgen_ldim%d .LT. kgen_mindim%d'%(rank+1, rank+1) }
                        ifmindim = part_append_genknode(ifeval, EXEC_PART, block_statements.IfThen, attrs=attrs)

                        attrs = {'variable': 'kgen_mindim%d'%(rank+1), 'sign': '=', 'expr': 'kgen_ldim%d'%(rank+1)}
                        part_append_genknode(ifmindim, EXEC_PART, statements.Assignment, attrs=attrs)

                        attrs = {'items': [ 'kgen_udim%d'%(rank+1) ], 'specs': ['UNIT = kgen_read_unit']}
                        part_append_genknode(ifeval, EXEC_PART, statements.Read, attrs=attrs)
                        bounds.append('kgen_ldim%d:kgen_udim%d'%(rank+1, rank+1))

                        attrs = {'expr': 'kgen_udim%d .GT. kgen_maxdim%d'%(rank+1, rank+1) }
                        ifmaxdim = part_append_genknode(ifeval, EXEC_PART, block_statements.IfThen, attrs=attrs)

                        attrs = {'variable': 'kgen_maxdim%d'%(rank+1), 'sign': '=', 'expr': 'kgen_udim%d'%(rank+1)}
                        part_append_genknode(ifmaxdim, EXEC_PART, statements.Assignment, attrs=attrs)

                    attrs = {'items': [ '%s(%s)'%(newvarstr, ', '.join(bounds)) ] , 'specs': ['UNIT = kgen_read_unit']}
                    part_append_genknode(ifeval, EXEC_PART, statements.Read, attrs=attrs)

            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    attrs = {'specs': ['UNIT=kgen_read_unit'], 'items': [ var.name ]}
                    part_append_genknode(ifeval, EXEC_PART, statements.Read, attrs=attrs)

        attrs = {'specs': ['UNIT=kgen_read_unit']}
        part_append_genknode(ifeval, EXEC_PART, statements.Close, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_evalinvoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_readsubp_evalinvoke_L%d + 1'%lineno}
        part_append_genknode(ifeval, EXEC_PART, statements.Assignment, attrs=attrs)

        # if warmup stage
        # - allocate an array and save data variable into the array
        attrs = {'expr': '.NOT. kgen_evalstage .AND. kgen_warmupstage .AND. .NOT. kgen_mainstage'}
        ifwarmup = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        idx += 1

        for vindex, (varstr, var) in enumerate(vars):
            attrs = {'expr': '.NOT. ALLOCATED(kgen_arr_%s_L%d)'%(var.name, lineno) }
            ifnoalloc = part_append_genknode(ifwarmup, EXEC_PART, block_statements.IfThen, attrs=attrs)

            if vindex == 0:
                attrs = {'variable': 'kgen_readsubp_maxinvoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_readsubp_evalinvoke_L%d + 1'%lineno}
                part_append_genknode(ifnoalloc, EXEC_PART, statements.Assignment, attrs=attrs)

            stmt = var.parent
            is_class_derived = check_class_derived(stmt)
            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    parts = varstr.split('%')
                    parts[-1] = var.name
                    newvarstr = '%'.join(parts)

                    bounds = []
                    for rank in range(var.rank):
                        bounds.append('kgen_mindim%d:kgen_maxdim%d'%(rank+1, rank+1))

                    attrs = {'items': ['kgen_arr_%s_L%d(kgen_readsubp_maxinvoke_L%d, %s)'%(var.name, lineno, lineno, ', '.join(bounds))]}
                    part_append_genknode(ifnoalloc, EXEC_PART, statements.Allocate, attrs=attrs)
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    pass
                attrs = {'items': ['kgen_arr_%s_L%d(kgen_readsubp_maxinvoke_L%d)'%(var.name, lineno, lineno)]}
                part_append_genknode(ifnoalloc, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_evalinvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_maininvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Assignment, attrs=attrs)

        l = [ 'kgen_mpirank', '"."', 'kgen_openmptid', '"."', 'kgen_kernelinvoke', '"."', 'kgen_readsubp_warmupinvoke_L%d'%lineno]
        attrs = {'specs': ['kgen_read_filepath_L%d'%lineno, 'FMT="(A,I0,A,I0,A,I0,A,I0)"' ], 'items': [ '"%s.L%d."'%(filename, lineno) ] + l}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Write, attrs=attrs)

        # file open
        attrs = {'specs': ['NEWUNIT=kgen_read_unit', 'FILE=kgen_read_filepath_L%d'%lineno, 'STATUS="OLD"', 'ACCESS="STREAM"', \
            'FORM="UNFORMATTED"', 'ACTION="READ"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Open, attrs=attrs)

        for varstr, var in vars:
            stmt = var.parent
            is_class_derived = check_class_derived(stmt)
            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    parts = varstr.split('%')
                    parts[-1] = var.name
                    newvarstr = '%'.join(parts)

                    bounds = []
                    for rank in range(var.rank):
                        attrs = {'items': [ 'kgen_ldim%d'%(rank+1) ], 'specs': ['UNIT = kgen_read_unit']}
                        part_append_genknode(ifwarmup, EXEC_PART, statements.Read, attrs=attrs)

                        attrs = {'items': [ 'kgen_udim%d'%(rank+1) ], 'specs': ['UNIT = kgen_read_unit']}
                        part_append_genknode(ifwarmup, EXEC_PART, statements.Read, attrs=attrs)
                        bounds.append('kgen_ldim%d:kgen_udim%d'%(rank+1, rank+1))

                    attrs = {'items': [ '%s(%s)'%(newvarstr, ', '.join(bounds)) ] , 'specs': ['UNIT = kgen_read_unit']}
                    part_append_genknode(ifwarmup, EXEC_PART, statements.Read, attrs=attrs)

                    attrs = {'variable': 'kgen_arr_%s_L%d(kgen_readsubp_warmupinvoke_L%d + 1, %s)'%(var.name, lineno, lineno, ', '.join(bounds)), \
                        'sign': '=', 'expr': '%s( %s )'%(newvarstr, ', '.join(bounds)) }
                    part_append_genknode(ifwarmup, EXEC_PART, statements.Assignment, attrs=attrs)

            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    attrs = {'specs': ['UNIT=kgen_read_unit'], 'items': [ var.name ]}
                    part_append_genknode(ifwarmup, EXEC_PART, statements.Read, attrs=attrs)

                attrs = {'variable': 'kgen_arr_%s_L%d(kgen_readsubp_warmupinvoke_L%d + 1)'%(var.name, lineno, lineno), 'sign': '=', 'expr': var.name}
                part_append_genknode(ifwarmup, EXEC_PART, statements.Assignment, attrs=attrs)

        # file close
        attrs = {'specs': ['UNIT=kgen_read_unit']}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Close, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_warmupinvoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_readsubp_warmupinvoke_L%d + 1'%lineno}
        part_append_genknode(ifwarmup, EXEC_PART, statements.Assignment, attrs=attrs)

        # if main stage
        # - copy data from array to data variable

        attrs = {'expr': '.NOT. kgen_evalstage .AND. .NOT. kgen_warmupstage .AND. kgen_mainstage'}
        ifmain = part_insert_genknode(node.kgen_parent, EXEC_PART, block_statements.IfThen, attrs=attrs, index=idx)
        idx += 1

        attrs = {'variable': 'kgen_readsubp_evalinvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifmain, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_warmupinvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifmain, EXEC_PART, statements.Assignment, attrs=attrs)

        for varstr, var in vars:
            stmt = var.parent
            is_class_derived = check_class_derived(stmt)
            if var.is_array():
                if is_zero_array(var, stmt): continue
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    parts = varstr.split('%')
                    parts[-1] = var.name
                    newvarstr = '%'.join(parts)

                    bounds = []
                    for rank in range(var.rank):
                        bounds.append('LBOUND(%s, %d):UBOUND(%s, %d)'%(newvarstr, rank+1, newvarstr, rank+1))

                    attrs = {'variable': newvarstr, 'sign': '=', 'expr': 'kgen_arr_%s_L%d(kgen_readsubp_maininvoke_L%d + 1, %s)'%(var.name, lineno, lineno, ', '.join(bounds))}
                    part_append_genknode(ifmain, EXEC_PART, statements.Assignment, attrs=attrs)
            else: # scalar
                if stmt.is_derived() or is_class_derived:
                    raise Exception('Derived type is not supported for manual state generation yet.')
                else: # intrinsic type
                    pass

                attrs = {'variable': var.name, 'sign': '=', 'expr': 'kgen_arr_%s_L%d(kgen_readsubp_maininvoke_L%d)'%(var.name, lineno, lineno)}
                part_append_genknode(ifmain, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_maininvoke_L%d'%lineno, 'sign': '=', 'expr': 'kgen_readsubp_maininvoke_L%d + 1'%lineno}
        part_append_genknode(ifmain, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'kgen_readsubp_maininvoke_L%d .GE. kgen_readsubp_maxinvoke_L%d'%(lineno,lineno) }
        ifmax = part_append_genknode(ifmain, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_readsubp_maininvoke_L%d'%lineno, 'sign': '=', 'expr': '0'}
        part_append_genknode(ifmax, EXEC_PART, statements.Assignment, attrs=attrs)

        part_insert_comment(node.kgen_parent, EXEC_PART, idx, '')
